#include <windows.h>
#include <ctype.h>              /* for isalnum() decl */
#include <string.h>             /* string calls for check_file_name() */
#include <io.h>                 /* for access() decl */
#include <stdlib.h>             /* for _ltoa() decl */
#include <stdio.h>              /* for fopen(), etc calls */
#include "wintel.h"

int FTP_access=0;   /* 0 = no access, 1 = restricted access, 2 = unrestricted access */

int GetAccessPaths(FTPCON far *pCon) {
    char far *lpRead,
         far *lpWrite,
         far *lpUser,
         far *lpPath,
         far *p;

    lpUser=(char far *)GlobalLock(pCon->hUser);
    if (!lpUser) {
        OutputDebugString("Couldn't lock down username!\r\n");
        return(0);
    }
    lpRead=(char far *)GlobalLock(pCon->hReadPath);
    if (!lpRead) {
        OutputDebugString("Couldn't lock down read pathname!\r\n");
        GlobalUnlock(pCon->hUser);
        return(0);
    }
    lpWrite=(char far *)GlobalLock(pCon->hWritePath);
    if (!lpWrite) {
        OutputDebugString("Couldn't lock down write pathname!\r\n");
        GlobalUnlock(pCon->hUser);
        GlobalUnlock(pCon->hReadPath);
        return(0);
    }
    GetPrivateProfileString("FTP Users",lpUser,"not found",lpRead,256,"ncsa_net.ini");
    if (!lstrcmp(lpRead,"not found"))
        GetPrivateProfileString("FTP Users","default",",",lpRead,256,"ncsa_net.ini");
    p=lpRead;
    while (*p && (*p != ','))
        p++;
    if (*p)
        *p++='\0';
    lstrcpy(lpWrite,p);
    lpPath=(char far *)GlobalLock(pCon->hPath);
    if (lpPath) {
        lstrcpy(lpPath,lpRead);
        if (!lpPath[0]) {
            GetPrivateProfileString("FTP Users","default",",",lpPath,256,"ncsa_net.ini");
            p=lpPath;
            while (*p && (*p != ','))
                p++;
            if (*p)
                *p++='\0';
            if (!lpPath[0])
                lstrcpy(lpPath,"c:\\");
        }
        GlobalUnlock(pCon->hPath);
    }
    GlobalUnlock(pCon->hUser);
    GlobalUnlock(pCon->hReadPath);
    GlobalUnlock(pCon->hWritePath);
}

int checkDirectoryAccess(FTPCON far *pCon, char far *lpPath, int iAccessType) {
    char far *lpAllowedPath;
    HGLOBAL hAllowedPath;

    switch (iAccessType) {
        case READ_ACCESS:
            hAllowedPath=pCon->hReadPath;
            break;
        case WRITE_ACCESS:
            hAllowedPath=pCon->hWritePath;
            break;
        default:
            hAllowedPath=NULL;
    }
    lpAllowedPath=(char far *)GlobalLock(hAllowedPath);
    if (!lpAllowedPath)
        return(0);
    if (!lstrncmp(lpAllowedPath,lpPath,lstrlen(lpAllowedPath))) {
        GlobalUnlock(hAllowedPath);
        return(1);
    }
    GlobalUnlock(hAllowedPath);
    return(0);
}

/****************************************************************************
*   encryptAndCompare ( ps, en )
*   Compute and check the encrypted password
*/
int encryptAndCompare(char far *lpToEncrypt, char far *lpEncryptedPassword) {
    int ck;
    char far *p,c;

    ck=0;
    p=lpToEncrypt;
    while (*p)                          /* checksum the string */
        ck+=(int)*p++;
    c=(char)ck;
    while (*lpEncryptedPassword) {
        if((((*lpToEncrypt^c)|32)&127)!=*lpEncryptedPassword)     /* XOR with checksum */ 
            return(0);
        if(*lpToEncrypt)
            lpToEncrypt++;
        else
            c++;                        /* increment checksum to hide length */
        lpEncryptedPassword++;
    }
    return(1);
}

/****************************************************************************
*   passwordCheck ( us, ps )
*   Check the password file for the user/password combination from ftp.
*   Returns valid/invalid  (1/0)
*/
int passwordCheck(char far *lpUser, char far *lpPass) {
    char buf[81], far *p;
    char passfile[80];
    FILE *fp;
    
    GetPrivateProfileString("FTP","password file","",passfile,80,"ncsa_net.ini");
    if(NULL==(fp=fopen(passfile,"r"))) {
        OutputDebugString("Unable to open password file.\r\n"); 
        return(0);
    }

    p=lpPass;               /* Remove the FTP protocol end-of-line chars */
    while (*p != '\015')
        p++;
    *p='\0';

    while (NULL != fgets(buf,80,fp)) {
        p=buf;
        while (*p != '\n')
            p++;
        *p='\0';                            /* remove \n */
        p=buf;
        while (*p != ':')
            p++;
        *p++='\0';
        if(!lstrcmp(buf,lpUser) && encryptAndCompare(lpPass,p)) {         /* does password check ?*/
            fclose(fp);
            return(1);
        }
    }
    fclose(fp);
    return(0);
}

/* check for required password for FTP access */
int accessCheck(char far *lpUser) {
    char buf[80];
    
    lstrcpy(buf,lpUser);
    AnsiLower(buf);
    if (!lstrcmp(buf,"anonymous")) {
        lstrcpy(lpUser,buf);
        GetPrivateProfileString("FTP","anonymous access","no",buf,80,"ncsa_net.ini");
        if (!lstrcmp(buf,"yes"))
            return(ANON_LOGIN);
        return(ANON_LOGIN_DENIED);
    }
    if (FTP_access==2)
        return(PASSWORD_NREQ);
    return(NEED_PASSWORD);      /* should have support for users w/o passwds here. */
}

/* should really free up the memory that this fn uses sometime. */
int TranslateAndWrite(HFILE hFile, char far *lpData,int iCount) {
    static HGLOBAL hData=NULL;
    char far *lpTmp;
    int iPos=0,
        idx,
        iTmp,
        iTmp2;

    if (hData==NULL) {
        hData=GlobalAlloc(GHND,32768);
        if (!hData) {
            OutputDebugString("Error! Could not allocate temp memory!\r\n");
            return(HFILE_ERROR);
        }
    }
    if ((lpTmp=(char far *)GlobalLock(hData))==NULL) {
        OutputDebugString("Error! Could not lock down temp memory!\r\n");
        return(HFILE_ERROR);
    }
    while (iPos<iCount) {
        iTmp=min(16384,iCount-iPos);
        _fmemcpy(lpTmp,&lpData[iPos],iTmp);
        iTmp2=0;
        while (iTmp2<iTmp) {
            if (((lpTmp[iTmp2]==0x0a)&&(iTmp==0))||((lpTmp[iTmp2]==0x0a)&&(lpTmp[iTmp2-1]!=0x0d))) {
                for (idx=iTmp; idx>iTmp2; idx--)  /* move remainder back a char to make room */
                    lpTmp[idx]=lpTmp[idx-1];
                lpTmp[iTmp2]=0x0d;
                iTmp++;     /* add one to total length */
                iTmp2++;    /* skip over the 0x0a that we just moved back a char */
            }
            iTmp2++;
        }
        if (_lwrite(hFile,lpTmp,iTmp)==(UINT)HFILE_ERROR) {
            OutputDebugString("Error writing to ASCII file!\r\n");
            return(HFILE_ERROR);
        }
        iPos+=iTmp;
    }
    return(1);
}

int WriteFileBlock(FTPCON far *pCon, int iEnd) {
    char far *lpData;
    int iTmp;
    WORD wDataIdx;

    if (pCon->hFile == HFILE_ERROR) {
        OutputDebugString("File handle is invalid!\r\n");
        return(0);
    }
    lpData=GlobalLock(pCon->hData);
    if (!lpData) {
        OutputDebugString("Error locking down data handle!\r\n");
        return(0);
    }
    wDataIdx=pCon->wDataIdx;
    while (pCon->wDataIdx) {
        iTmp=(int)(min(pCon->wDataIdx,(WORD)32767));
        if (pCon->mode==FASCII) {
            if (TranslateAndWrite(pCon->hFile,lpData,iTmp)==(UINT)HFILE_ERROR) {
                OutputDebugString("Error writing to file!\r\n");
                return(0);
            }
        } else {
            if (_lwrite(pCon->hFile,lpData,iTmp)==(UINT)HFILE_ERROR) {
                OutputDebugString("Error writing to file!\r\n");
                return(0);
            }
        }
        pCon->wDataIdx-=iTmp;
        lpData+=iTmp;
    }
    pCon->dwFileIdx+=wDataIdx;
    pCon->wDataIdx=0;
    GlobalUnlock(pCon->hData);
    if (iEnd) {
//        OutputDebugString("Must be done reading from net, closing file.\r\n");
        _lclose(pCon->hFile);
        pCon->hFile=HFILE_ERROR;
        pCon->wDataIdx=pCon->wDataLen=0;
        return(1);
    } else {
        pCon->wDataIdx=pCon->wDataLen=0;
        return(1);
    }
}

int ReadFileBlock(FTPCON far *pCon) {
    char far *lpData;

    pCon->wDataIdx=pCon->wDataLen=0;
    if (pCon->hFile == HFILE_ERROR)
        return(0);
    lpData=GlobalLock(pCon->hData);
    if (!lpData) {
        OutputDebugString("Error locking down data handle!\r\n");
        return(0);
    }
    pCon->wDataLen=_lread(pCon->hFile,lpData,DATA_BUF_SIZE);
    pCon->wDataIdx=0;
    pCon->dwFileIdx+=pCon->wDataLen;
    GlobalUnlock(pCon->hData);
    if (pCon->wDataLen < DATA_BUF_SIZE) {
//        OutputDebugString("Closing file.\r\n");
        _lclose(pCon->hFile);
        pCon->hFile=HFILE_ERROR;
        return(0);
    } else
        return(1);
}

int ReadASCIIFileBlock(FTPCON far *pCon) {
    static HGLOBAL hBlock=NULL;
    char far *lpBlock;
    char far *lpIndexer;
    char far *lpData;
    WORD wDataLen, wIdx, wDataIdx=0;
    int iEndOfFile=0;

    pCon->wDataIdx=pCon->wDataLen=0;
    if (pCon->hFile == HFILE_ERROR)
        return(0);
    lpData=GlobalLock(pCon->hData);
    if (!lpData) {
        OutputDebugString("Error locking down data handle!\r\n");
        return(0);
    }
    if (!hBlock)
        hBlock=GlobalAlloc(GHND,DATA_BUF_SIZE);
    lpBlock=(char far *)GlobalLock(hBlock);
    if (!lpData) {
        OutputDebugString("Error locking down tmp block handle!\r\n");
        return(0);
    }
    while ((DATA_BUF_SIZE - wDataIdx)&&(!iEndOfFile)) {
        wDataLen=_lread(pCon->hFile,lpBlock,DATA_BUF_SIZE-wDataIdx);
        if (wDataLen != (DATA_BUF_SIZE-wDataIdx))
            iEndOfFile=1;
        lpIndexer=lpBlock;
        for (wIdx=0; wIdx<wDataLen; wIdx++) {
            if (*lpIndexer != 0x0d) {
                *lpData++=*lpIndexer;
                wDataIdx++;
            }
            lpIndexer++;
        }
    }
    pCon->wDataLen=wDataIdx;
    pCon->wDataIdx=0;
    pCon->dwFileIdx+=pCon->wDataLen;
    GlobalUnlock(pCon->hData);
    if (iEndOfFile) {
//        OutputDebugString("Closing file.\r\n");
        _lclose(pCon->hFile);
        pCon->hFile=HFILE_ERROR;
        return(0);
    } else
        return(1);
}

int lstrncmp(char far *str1, char far *str2, int n) {
    while (n) {
        if (*str1 < *str2)
            return(-1);
        if (*str1 > *str2)
            return(1);
        if (*str1==0)
            return(0);
        str1++;
        str2++;
        n--;
    }
    return(0);
}

int FileType(char far *filename) {
    if(!lstrcmp(filename,"."))
        return(IS_DIRECTORY);
    if(!lstrcmp(filename,"\\"))
        return(IS_DIRECTORY);
    if (!filename[0])
        return(DOESNT_EXIST);
    if (n_findfirst(filename,0x10))
        return(0);
    if (!(dtaptr->att&0x10))    /* is a normal file */
        return(IS_FILE);
    return(IS_DIRECTORY);
}

void StripLastFromPath(char far *lpPath) {
    int i;

    i=lstrlen(lpPath);
    while ((lpPath[i] != '\\')&&(i>0)&&(lpPath[i] != ':'))
        i--;
    if (i) {
        if (lpPath[i-1]==':')
            i--;
        if (lpPath[i]==':') {
            lpPath[i+1]='\\';
            lpPath[i+2]=0;
        } else
            lpPath[i]=0;
    } else {
        lpPath[0]=0;
    }
}

int MungeNameToFullPathName(FTPCON far *pCon,char far *fname) {    int iLen;
    char strFile[256];
    char far *strDir;

    if (!fname) {
        OutputDebugString("No filename given!\r\n");
        return(0);
    }
    strDir=(char far *)GlobalLock(pCon->hPath);
    if (!strDir) {
        OutputDebugString("Couldn't lock path down.\r\n");
        return(0);
    }
    AnsiLower(fname);
    iLen=lstrlen(fname);
    while (((!isalnum(fname[iLen]))&&(fname[iLen] != '.')&&
            (fname[iLen] != '*')&&(fname[iLen]!='\\')&&(fname[iLen]!=':'))
            &&(iLen)) {
        fname[iLen]=0;
        iLen--;
    }
    if (fname[1]==':') {        /* Filename includes drive spec & full path */
        return(1);
    }
    if (fname[0]=='\\') {       /* Filename includes full path, no drive spec */
        strFile[0]=strDir[0];
        strFile[1]=':';
        lstrcpy(&strFile[2],fname);
        GlobalUnlock(pCon->hPath);
        lstrcpy(fname,strFile);
        return(1);
    }
    lstrcpy(strFile,strDir);
    GlobalUnlock(pCon->hPath);
    if (!lstrncmp(fname,"..",2)) {
        StripLastFromPath(strFile);
        fname+=3;
    }
    iLen=lstrlen(strFile);
    if (strFile[iLen-1]!='\\')
        lstrcat(strFile,"\\");
    lstrcat(strFile,fname);
    lstrcpy(fname,strFile);
    return(1);
}

int CreateDirList(FTPCON far *pCon, char far *dirname, int lstype) {
    WORD wTmp=0;
    char far *lpData;
    int iLen;

//    OutputDebugString("In CreateDirList.\r\n");
    MungeNameToFullPathName(pCon,dirname);
    if (!checkDirectoryAccess(pCon, dirname, READ_ACCESS))
        return(-1);

//    OutputDebugString("1\r\n");
    if ((lpData=GlobalLock(pCon->hData))==NULL) {
        OutputDebugString("Could not lock down data pointer!\r\n");
        return(0);
    }
//    OutputDebugString("2\r\n");
    iLen=lstrlen(dirname);
    while (iLen && !isalnum(dirname[iLen]) && (dirname[iLen] != '\\') &&
           (dirname[iLen] != '*') && (dirname[iLen] != '.'))
        dirname[iLen--]=0;
    if (dirname[iLen] == '\\')
        lstrcat(dirname,"*.*");
//    OutputDebugString("3\r\n");
    firstname(dirname,lstype,lpData);
//    OutputDebugString(lpData);
//    OutputDebugString("\r\n");
    do {
        lstrcat(lpData,"\015\012");
        wTmp+=lstrlen(&lpData[wTmp]);
//        OutputDebugString(&lpData[wTmp]);
//        OutputDebugString("\r\n");
    } while (nextname(lstype,&lpData[wTmp]));
    lpData[wTmp]=0;
    pCon->wDataLen=wTmp;
    pCon->wDataIdx=0;
    GlobalUnlock(pCon->hData);
//    OutputDebugString("4\r\n");
    return(1);
}

/**********************************************************************/
/* firstname
*  find the first name in the given directory which matches the wildcard
*  specification
*
*  expand '*' (unix) to '*.*' (dos)
*/

char savepath[256]; /* allocate enough room for the pathname, filename, & extension, plus a little bit of slush */
int rootlen;

int firstname(char far *path, int type, char far *ret) {
    int i,len;
    char far *p, far *q;

    if(!*path)
        return(0);

    len=lstrlen(path);
    i=0;
    rootlen=0;
    q=(char far *)savepath;
    p=path;
    while(*q=*p) {              /* basic string copy with extras */
        if(*p=='\\')
            rootlen=i+1;            /* rootlen = position of last \ */
        p++;
        q++;
        i++;
    }
    if(savepath[len-1]=='*' && rootlen==len-1) {
        savepath[len++]='.';
        savepath[len++]='*';
        savepath[len++]='\0';
    }
    switch(type) {
        case 1:
            if(n_findfirst(savepath,0x10))
                return(0);
             break;

        default: 
            if(n_findfirst(savepath,0x00))
                return(0);
             break;

      }
/*
*  copy file name, translating to lower case
*/
    q=ret;
    p=dtaptr->name;
    while(*p) {
        if(*p>='A' && *p<='Z')
            *q++= (*p++) + (char) 32;
        else
            *q++=*p++;
    }
/*
*  if it is a directory then put <DIR> after it
*/
    if(type) {
        p=ret+20;
        for(; q!=p; *q++=' ');      
        if(dtaptr->att&0x10) {
            *q++=' ';
            *q++='<';
            *q++='D';
            *q++='I';
            *q++='R';
            *q++='>';
        } else {
            wsprintf(q,"%8ld",dtaptr->size);
            return(lstrlen(ret));
        }
    }
    *q='\0';
    return(lstrlen(ret));
}

/**********************************************************************/
/* nextname
*  modify the path spec to contain the next file name in the
*  sequence as given by DOS
*
*  if at the end of the sequence, return NULL
*/
int nextname(int type, char far *ret) {
    char far *p, far *q;

    if(n_findnext())    /* check if there are any more filenames */
        return(NULL);
/*
*  copy file name, translate to lower case 
*/
    q=ret;
    p=dtaptr->name;
    while(*p) {
        if(*p >='A' && *p<='Z')
            *q++ = (*p++) + (char)32;
        else
            *q++=*p++;
    }
/*
* if it is a directory, then put <DIR> after it
*/
    if(type) {
        p=ret+20;
        for(; q!=p; *q++=' ');      
        if(dtaptr->att&0x10) {
            *q++=' ';
            *q++='<';
            *q++='D';
            *q++='I';
            *q++='R';
            *q++='>';
        } else {
            wsprintf(q,"%8ld",dtaptr->size);
            return(lstrlen(ret));
        }
    }
    *q='\0';
    return(lstrlen(ret));
}

/************************************************************************/
/*                             check_file_name                          */
/*                                                                      */
/*  A function to check for and enforce MS-DOS file name compatibility  */
/*  and uniqueness.  All non-MS-DOS filename characters from 'in_fname' */
/*  are deleted.  If the filename exists in the current directory, a    */
/*  unique name is created to avoid file name collision.                */
/*  'out_fname' must be large enough for the full filename (13 bytes).  */
/*                                                                      */
/*       Returns:    pointer to 'out_fname'                             */
/*                                                                      */
/*       Allen W. Todd                        Dec 24, 1990              */
/*       Rewritten for Windows Telnet-  Chris Wilson, 4/24/93           */
/************************************************************************/
char far *check_file_name(FTPCON far *pCon,char far *in_fname) {
    char buf[80],
        far *s,
        far *fname,
        ext_buf[4],
        far *fextn,
        num_buf[20],
        far *lpPath,
        tmp_buf[80],
        far *tmp_ptr;
    long count=0L;
    int iTmp;

    lpPath=(char far *)GlobalLock(pCon->hPath);
    if (!lpPath) {
        OutputDebugString("Couldn't lock path down.\r\n");
        return(NULL);
    }
    /* Make ready to copy in name */
    lstrcpy(tmp_buf,lpPath);
    if (tmp_buf[lstrlen(tmp_buf)-1] != '\\')
        lstrcat(tmp_buf,"\\");
    tmp_ptr=&tmp_buf[lstrlen(tmp_buf)];
    GlobalUnlock(pCon->hPath);

    lstrcpy(buf,in_fname);    /* make local copy */
    AnsiLower(buf);           /* convert to lower case */
    iTmp=lstrlen(buf);
    while ((buf[iTmp]<20)&&(iTmp))
        buf[iTmp--]='\0';
    while((s=_fstrpbrk(buf, "\"*+,/:;<=>?[\\]^| "))!=NULL)       // added space -CGW
        lstrcpy(s,s+1);       /* delete non-DOS chars */
    fname=&buf[0];            /* separate filename from file extension */
    if((fextn=_fstrchr(buf,'.'))!=NULL)
        *fextn++='\0';
    if(lstrlen (fname)>8)
        *(fname+8)='\0';   /* fix filename @ 8 char */
    if(lstrlen(fextn)>3)
        *(fextn+3)='\0';   /* fix file extension @ 3 char */
    lstrcpy(in_fname,fname);     /* build complete filename */
    if(fextn) {
        lstrcat(in_fname,".");
        lstrcat(in_fname,fextn);
    }
    lstrcpy(tmp_ptr,in_fname);      /* add in path to filename */
    while(_access(tmp_buf,0)==0) {  /* if file already exists */
        ext_buf[0]='\0';
        _ltoa(count++,num_buf,10);
        lstrcpy(in_fname,fname);    /* build complete filename */
        if(lstrlen(num_buf)+lstrlen(fextn) <= 3) {
            lstrcpy(ext_buf,fextn);
            lstrcat(ext_buf,num_buf);
        } else {
            if(lstrlen(num_buf)+lstrlen(fname) <= 8)
                lstrcat(in_fname,num_buf);
              else
                lstrcpy(in_fname+(8-lstrlen(num_buf)),num_buf);
        }
        if(ext_buf[0]) {
            lstrcat(in_fname,".");
            lstrcat(in_fname,ext_buf);
        }
        lstrcpy(tmp_ptr,in_fname);
    }
    lstrcpy(in_fname,tmp_buf);      /* Copy full pathname to filename */
    return(in_fname);
}




