/****************************************************************************

    PROGRAM: wintel.c

    PURPOSE: Windows networking kernel - Telnet & FTP

    FUNCTIONS:

    WinMain() - calls initialization function, processes message loop
    InitApplication() - initializes window data and registers window
    InitInstance() - saves instance handle and creates main window
    MainWndProc() - processes messages
    About() - processes messages for "About" dialog box

    COMMENTS:

        Windows can have several copies of your application running at the
        same time.  The variable hInst keeps track of which instance this
        application is so that processing will be to the correct window.

****************************************************************************/

#include <windows.h>            /* required for all Windows applications */
#include <string.h>
#include "wintel.h"            /* specific to this program          */
/*
#include "..\include\netevent.h"
#include "..\include\hostform.h"
#include "..\include\externs.h"
*/

HANDLE hInst;   /* current instance      */
HWND hWnd;      /* Main window handle.   */
HMENU  hMenu;
CONFIG FAR *tmpConfig;
GLOBALHANDLE hGlobalMem,hTitleMem;
BOOL bConfigure;                     
HCONNECTION ConnectionList=NULL;
int TempTest;
int pnum;
int debug=1;
char strTmp[512];
int yPos=0;
int NumConnections=0;

/****************************************************************************

    FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)

    PURPOSE: calls initialization function, processes message loop

    COMMENTS:

        Windows recognizes this function by name as the initial entry point 
        for the program.  This function calls the application initialization 
        routine, if no other instance of the program is running, and always 
        calls the instance initialization routine.  It then executes a message 
        retrieval and dispatch loop that is the top-level control structure 
        for the remainder of execution.  The loop is terminated when a WM_QUIT 
        message is received, at which time this function exits the application 
        instance by returning the value passed by PostQuitMessage(). 

        If this function must abort before entering the message loop, it 
        returns the conventional value NULL.  

****************************************************************************/

int PASCAL WinMain(hInstance, hPrevInstance, lpCmdLine, nCmdShow)
HANDLE hInstance;                /* current instance             */
HANDLE hPrevInstance;            /* previous instance            */
LPSTR lpCmdLine;                 /* command line                 */
int nCmdShow;                    /* show-window type (open/icon) */
{
    MSG msg;                     /* message                      */

    if (hPrevInstance) {
        MessageBox(NULL,"Already running TCP/IP kernel!",NULL,MB_OK);
        return(FALSE);
    }        
    if (!InitApplication(hInstance)) /* Initialize shared things */
        return (FALSE);      /* Exits if unable to initialize    */

    /* Perform initializations that apply to a specific instance */

    if (!InitInstance(hInstance, nCmdShow))
        return (FALSE);

    /* Acquire and dispatch messages until a WM_QUIT message is received. */

    while (GetMessage(&msg, NULL, NULL, NULL)) {
        TranslateMessage(&msg); /* Translates virtual key codes  */
        DispatchMessage(&msg);  /* Dispatches message to window  */
    }
    return (msg.wParam);    /* Returns the value from PostQuitMessage */
}


/****************************************************************************

    FUNCTION: InitApplication(HANDLE)

    PURPOSE: Initializes window data and registers window class

    COMMENTS:

        This function is called at initialization time only if no other 
        instances of the application are running.  This function performs 
        initialization tasks that can be done once for any number of running 
        instances.  

        In this case, we initialize a window class by filling out a data 
        structure of type WNDCLASS and calling the Windows RegisterClass() 
        function.  Since all instances of this application use the same window 
        class, we only need to do this when the first instance is initialized.  


****************************************************************************/

BOOL InitApplication(hInstance)
HANDLE hInstance;                  /* current instance       */
{
    WNDCLASS  wc;

    /* Fill in window class structure with parameters that describe the       */
    /* main window.                                                           */

    wc.style = CS_HREDRAW|CS_VREDRAW;                    /* Class style(s).                    */
    wc.lpfnWndProc = MainWndProc;       /* Function to retrieve messages for  */
                                        /* windows of this class.             */
    wc.cbClsExtra = 0;                  /* No per-class extra data.           */
    wc.cbWndExtra = 0;                  /* No per-window extra data.          */
    wc.hInstance = hInstance;           /* Application that owns the class.   */
    wc.hIcon = LoadIcon(hInstance, "NCSA");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH); 
    wc.lpszMenuName =  NULL;   /* Name of menu resource in .RC file. */
    wc.lpszClassName = "wintelWClass"; /* Name used in call to CreateWindow. */

    /* Register the window class and return success/failure code. */
    return (RegisterClass(&wc));


}


/****************************************************************************

    FUNCTION:  InitInstance(HANDLE, int)

    PURPOSE:  Saves instance handle and creates main window

    COMMENTS:

        This function is called at initialization time for every instance of 
        this application.  This function performs initialization tasks that 
        cannot be shared by multiple instances.  

        In this case, we save the instance handle in a static variable and 
        create and display the main program window.  
        
****************************************************************************/

BOOL InitInstance(hInstance, nCmdShow)
    HANDLE          hInstance;          /* Current instance identifier.       */
    int             nCmdShow;           /* Param for first ShowWindow() call. */
{
                  
    int xScreen, yScreen;
    FARPROC lpMenu;
    
    /* Save the instance handle in static variable, which will be used in  */
    /* many subsequence calls from this application to Windows.            */

    hInst = hInstance;

    /* Create a main window for this application instance.  */
    xScreen=GetSystemMetrics(SM_CXSCREEN);
    yScreen=GetSystemMetrics(SM_CYSCREEN);
    lpMenu=MakeProcInstance((FARPROC)MainWndProc,hInst);

    hWnd = CreateWindow(
        "wintelWClass",                /* See RegisterClass() call.          */
        "TCPWin",                       /* Text for window title bar.         */
        WS_SYSMENU,                     /* Window style.                      */
        xScreen/3,                      /* Default horizontal position.       */
        yScreen/3,                      /* Default vertical position.         */
        xScreen/3,                      /* Default width.                     */
        yScreen/3,                      /* Default height.                    */
        NULL,                           /* Overlapped windows have no parent. */
        NULL,                           /* Use the window class menu.         */
        hInstance,                      /* This instance owns this window.    */
        NULL                            /* Pointer not needed.                */
    );

    /* If window could not be created, return "failure" */
    hMenu= GetSystemMenu(hWnd,FALSE);
    AppendMenu(hMenu,MF_SEPARATOR,0,NULL);
    AppendMenu(hMenu,MF_STRING,IDM_ABOUT,"About NCSA TCPWin Kernel...");
    AppendMenu(hMenu,MF_STRING,IDM_OPENTELNETDLG,"Open Telnet Connection...");
    
    
    if (!hWnd)
        return (FALSE);

    /* Make the window visible; update its client area; and return "success" */

    ShowWindow(hWnd, SW_SHOWMINIMIZED);  /* Show the window   */
    UpdateWindow(hWnd);          /* Sends WM_PAINT message      */

    {
        WSADATA wsaData;
        WSAStartup(0x0101,&wsaData);    /* Initialize the network */
    }
    SetTimer(hWnd,1,2,NULL);        

#ifdef FTP
    InitFTPListen();
#endif
    return (TRUE);
}

/****************************************************************************

    FUNCTION: MainWndProc(HWND, UINT, WPARAM, LPARAM)

    PURPOSE:  Processes messages

    MESSAGES:

    WM_COMMAND    - application menu (About dialog box)
    WM_DESTROY    - destroy window

****************************************************************************/

long FAR PASCAL MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    HCONNECTION hCon;
    CONNECTION far *con=NULL;
    CONNECTION far *pNewCon=NULL;
    HSCREEN hgScr;
    SCREEN far *fpScr;
    HGLOBAL hBuffer;
    LPSTR lpBuffer; 
    static BOOL bTimerSemaphore=FALSE;
    int tmp=0,c,cnt;
    unsigned char buf[512];
    HCONNECTION hTmp,hNewCon;
    int ret,namelen;
    struct sockaddr_in name;
        
    
    switch (message) {

    case WM_SYSCHAR:
        switch (wParam) {
            case 'a':
            case 'o':
                PostMessage(hWnd,WM_SYSCOMMAND,IDM_OPENTELNETDLG,NULL);
                break;
            default:
                return (DefWindowProc(hWnd, message, wParam, lParam));
                break;
        }
        break;
                    
    case WM_SYSCOMMAND:       /* message: command from application menu */
        switch(wParam) {
            case IDM_ABOUT:
                DoDialog("ABOUTBOX",About);
                break;
            case IDM_OPENTELNETDLG:
                OpenTelnetConnection();                                
                break;
            
        }
        return (DefWindowProc(hWnd, message, wParam, lParam));

    case WM_MYSCREENCHANGEBKSP:
        hCon=FindConnectionFromScreen((HSCREEN)lParam);
        if (hCon==NULL) {
            OutputDebugString("Couldn't get hCon from screen! \n\r");
            break;
        }
        con=(CONNECTION far *) GlobalLock(hCon);
        if (con==NULL) {
            OutputDebugString("Couldn't lock down con \n\r");
            break;
        }
        con->backspace=wParam;
        if (con->backspace==VK_BACK) {
//            OutputDebugString("CHANGED to BKSP \r\n");
            con->ctrl_backspace=0x7f;
        } else {    
//            OutputDebugString("CHANGED to DELETE \r\n");
            con->ctrl_backspace=VK_BACK;
        }    
        GlobalUnlock(hCon);
        break;
        
    case WM_MYSCREENCHAR:
        hCon=FindConnectionFromScreen((HSCREEN)lParam);
        if (hCon==NULL) {
            OutputDebugString("Couldn't get hCon from screen! \n\r");
            break;
        }
        con=(CONNECTION far *) GlobalLock(hCon);
        if (con==NULL) {
            OutputDebugString("Couldn't lock down con \n\r");
            break;
        }
        if (wParam==VK_BACK)
            wParam=con->backspace;
          else
            if (wParam==0x7f)
                wParam=con->ctrl_backspace;
        send(con->socket, (char FAR *)&wParam,1,NULL);
        GlobalUnlock(hCon);
        break;
    
    case WM_MYSYSCHAR:
        c= (char) wParam;
        hCon=FindConnectionFromScreen((HSCREEN)lParam);
        if (hCon==NULL) {
            OutputDebugString("Couldn't get hCon from screen! \n\r");
            break;
        }
        con=(CONNECTION far *) GlobalLock(hCon);
        if (con==NULL) {
            OutputDebugString("Couldn't lock down con \n\r");
            break;
        }
        switch (c) {
            case 'f':
                getsockname(con->socket,(struct sockaddr far *)&name,(int FAR *)&namelen);
                wsprintf(buf,"ftp %d.%d.%d.%d\n",
                    name.sin_addr.S_un.S_un_b.s_b1,
                    name.sin_addr.S_un.S_un_b.s_b2,
                    name.sin_addr.S_un.S_un_b.s_b3,
                    name.sin_addr.S_un.S_un_b.s_b4);
//                OutputDebugString(buf);
                send(con->socket,buf,lstrlen(buf),NULL);
                break;                
            case 'a':
            case 'o':
                PostMessage(hWnd,WM_SYSCOMMAND,IDM_OPENTELNETDLG,NULL);
                break;
            case 'x':
                hgScr=con->hScreen;
                fpScr=(SCREEN far *)GlobalLock(hgScr);
                if (fpScr == NULL) OutputDebugString("Couldn't lock screen.\r\n");
                PostMessage(fpScr->hWnd,WM_CLOSE,NULL,NULL);
                break;
                                
            case 'b':
                if (con->prev) {
                    hNewCon=con->prev;
                    pNewCon=(CONNECTION far *) GlobalLock(hNewCon);
                    if (pNewCon==NULL) {
                        OutputDebugString("Couldn't lock down con \n\r");
                        break;
                    }
                    hgScr=pNewCon->hScreen;
                    fpScr=(SCREEN far *)GlobalLock(hgScr);
                    if (fpScr == NULL) OutputDebugString("Couldn't lock screen.\r\n");
                    SetActiveWindow(fpScr->hWnd);
                    GlobalUnlock(hgScr);
                    GlobalUnlock(hNewCon);                    
                }    
                else { /* front of list - find end */
                    hNewCon=ConnectionList;
                    while (hNewCon) {
                        pNewCon=(CONNECTION far *) GlobalLock(hNewCon);
                        if (pNewCon==NULL) {
                            OutputDebugString("Couldn't lock down con \n\r");
                            break;
                        }
                        if (!(pNewCon->next)) break;
                        hTmp=pNewCon->next;
                        GlobalUnlock(hNewCon);
                        hNewCon=hTmp;
                    }                            
                    hgScr=pNewCon->hScreen;
                    fpScr=(SCREEN far *)GlobalLock(hgScr);
                    if (fpScr == NULL) OutputDebugString("Couldn't lock screen.\r\n");
                    SetActiveWindow(fpScr->hWnd);
                    GlobalUnlock(hgScr);
                    GlobalUnlock(hNewCon);                    
                }
                break;
            case 'n':
                if (con->next) {
                    hNewCon=con->next;
                    pNewCon=(CONNECTION far *) GlobalLock(hNewCon);
                    if (pNewCon==NULL) {
                        OutputDebugString("Couldn't lock down con \n\r");
                        break;
                    }
                    hgScr=pNewCon->hScreen;
                    fpScr=(SCREEN far *)GlobalLock(hgScr);
                    if (fpScr == NULL) OutputDebugString("Couldn't lock screen.\r\n");
                    SetActiveWindow(fpScr->hWnd);
                    GlobalUnlock(hgScr);
                    GlobalUnlock(hNewCon);                    
                }
                else { /* end of list - wrap */
                    hNewCon=ConnectionList;
                    pNewCon=(CONNECTION far *) GlobalLock(hNewCon);
                    if (pNewCon==NULL) {
                        OutputDebugString("Couldn't lock down con \n\r");
                        break;
                    }
                    hgScr=pNewCon->hScreen;
                    fpScr=(SCREEN far *)GlobalLock(hgScr);
                    if (fpScr == NULL) OutputDebugString("Couldn't lock screen.\r\n");
                    SetActiveWindow(fpScr->hWnd);
                    GlobalUnlock(hgScr);
                    GlobalUnlock(hNewCon);                    
                }                
                break;
        }                
        GlobalUnlock(hCon);
        break;
            
    case WM_MYSCREENBLOCK:
        hBuffer= (HGLOBAL) wParam;
        lpBuffer= GlobalLock(hBuffer);
    
        hCon=FindConnectionFromScreen((HSCREEN)lParam);
        if (hCon==NULL) {
            OutputDebugString("Couldn't get hCon from screen! \n\r");
            break;
        }
        con=(CONNECTION far *) GlobalLock(hCon);
        if (con==NULL) {
            OutputDebugString("Couldn't lock down con \n\r");
            break;
        }
        send(con->socket,lpBuffer,lstrlen(lpBuffer),NULL);
        
        GlobalUnlock(hCon);
        GlobalUnlock(hBuffer);
        break;

    case WM_MYSCREENCLOSE:
        hCon=FindConnectionFromScreen((HSCREEN)lParam);
        if (hCon==NULL) {
            OutputDebugString("Couldn't get hCon from screen! \n\r");
            break;
        }
        con=(CONNECTION far *) GlobalLock(hCon);
        if (con==NULL) {
            OutputDebugString("Couldn't lock down con \n\r");
            break;
        }
        closesocket(con->socket);
        DeleteConnectionFromList(con->socket);
        GlobalUnlock(hCon);
        GlobalFree(hCon);    
        break;        

    case WM_QUERYOPEN:
        return(0);
        break;

    case WM_DESTROY:          /* message: window being destroyed */
        bTimerSemaphore=TRUE;
//        OutputDebugString("wintel:WM_DESTROY");
        hCon=ConnectionList;
#ifdef OLD_WAY
        for (iTmp=0; iTmp < NumConnections; iTmp++) {
            if (hCon==NULL) break;
#else
        while (hCon) {
#endif                        
            con=(CONNECTION far *) GlobalLock(hCon);
            if (con==NULL) {
                OutputDebugString("Unable to lock down connection list!\r\n");
                break;
            }
            closesocket(con->socket);
            wsprintf(strTmp,"wintel: closed socket #%d \r\n",con->socket);
//            OutputDebugString(strTmp);
            hTmp=con->next;
            GlobalUnlock(hCon);
            GlobalFree(hCon);
            hCon=hTmp;
        }
        ConnectionList=NULL;
        CloseFTP();
        PostQuitMessage(0);
        break;

    case WM_TIMER:
        if (!bTimerSemaphore) { /* Keeps us from going re-entrant */
            unsigned long count=0;

            bTimerSemaphore=TRUE;
#ifdef FTP
            /* Deal with any pending FTP data or new connections */
            ProcessFTPConnections();
#endif

            /* Deal with any pending data for open Telnet sessions */
            hCon=ConnectionList;
            while (hCon) {
                con=(CONNECTION far *) GlobalLock(hCon);
                if (con==NULL) {
                    OutputDebugString("Unable to lock down connection list!\r\n");
                    break;
                }    
                ret=ioctlsocket((SOCKET)con->socket,(long)FIONREAD,(u_long FAR *)&count);
                if (ret==SOCKET_ERROR) { /* closed? */
                    fpScr= (SCREEN far *) GlobalLock(con->hScreen);
                    if (fpScr==NULL) OutputDebugString("fpScr is NULL! \r\n");
                    DestroyWindow(fpScr->hWnd);
                    NumConnections--;
                    GlobalUnlock(con->hScreen);
                    hTmp=con->next;
                    DeleteConnectionFromList(con->socket);
                    GlobalUnlock(hCon);
                    GlobalFree(hCon);                    
                    break;                    
                } 
                if (count) {
                    cnt=recv(con->socket,buf,512,NULL);
//                    if (!con->echo) send(con->socket,buf,cnt,NULL);
                    buf[cnt]=0;
                    parse((CONNECTION far *)con,(unsigned char far *)buf,(int)lstrlen(buf));
                    ScreenEm(buf,cnt,con->hScreen);
                }
                hTmp=con->next;
                GlobalUnlock(hCon);
                hCon=hTmp;
            }
            bTimerSemaphore=FALSE;
        }    

    default:              /* Passes it on if unproccessed    */
        return (DefWindowProc(hWnd, message, wParam, lParam));
    }

    return (NULL);
}


HCONNECTION FindConnectionFromScreen(HSCREEN hScreen) {
    int idx;
    HCONNECTION hCon,hTmp;
    CONNECTION far *con=NULL;

    hCon=ConnectionList;
    con=(CONNECTION far *) GlobalLock(hCon);
    if (con==NULL) OutputDebugString("Unable to lock down connection list!\r\n");
    
    for (idx=0; idx< MAX_NUM_CONNECTIONS ; idx++) 
        if (con->hScreen == hScreen) {
            GlobalUnlock(hCon);
            return(hCon);
        } else {
            hTmp=con->next;
            GlobalUnlock(hCon);
            hCon=hTmp;
            if (hCon==ConnectionList) return(NULL);
            con=(CONNECTION far *) GlobalLock(hCon);
            if (con==NULL) return (NULL);
        }    
            
    return (NULL);        
}            

int OpenTelnetConnection(void) {
    int nReturn,telstate,ret;
    struct sockaddr_in         sockaddr;
    struct sockaddr_in         remote_addr;
    struct hostent          far *remote_host;
    CONNECTION far *con=NULL;
    char buf[128],tmpName[128],far * tmpCommaLoc;
    SCREEN far *fpScr;
    SOCKET s;
    int iHostNum=0;
        
    bConfigure=FALSE;
    hGlobalMem= GlobalAlloc(GPTR,sizeof(CONFIG));
    tmpConfig= (CONFIG FAR *) GlobalLock(hGlobalMem);
    
    nReturn=DoDialog("OPENTELNETDLG",OpenTelnetDlg);
    if (nReturn==FALSE) return(FALSE);
    if (bConfigure) {
//        OutputDebugString("About to do dialog! \r\n");
        DoDialog("CONFIG_DLG",ConfigSessionDlg);   
    }
    telstate=GetNewConnection();
    if (telstate==-1) return(0);
    
    con=(CONNECTION far *) GlobalLock(ConnectionList);
    if (con==NULL) OutputDebugString("Unable to lock down connection list!\r\n");

    for(;;) {
        char tmpBuf[80];
        
        wsprintf(tmpBuf,"host%d",iHostNum++);
        GetPrivateProfileString("hosts",tmpBuf,"",tmpName,128,"ncsa_net.ini");
        tmpCommaLoc=_fstrchr(tmpName,',');  
        if (tmpCommaLoc) *tmpCommaLoc='\0';
        if (tmpName)
            if (!lstrcmp(tmpName,tmpConfig->title)) break;
            else continue;
        else break; 
    }
    tmpCommaLoc++;
    if (tmpName) 
        if (!lstrcmp(tmpCommaLoc,"backspace")) con->backspace=VK_BACK;
            else con->backspace=0x7f;
    else con->backspace=VK_BACK;        
    
    if (con->backspace==0x7f) {
        con->ctrl_backspace=0x08;
        tmpConfig->backspace=FALSE;
    }    
    else {
        tmpConfig->backspace=TRUE;
        con->ctrl_backspace=0x7f;
    }
    con->hScreen=(HSCREEN)InitNewScreen((CONFIG FAR *)tmpConfig);
    con->telstate=OPEN_STANDBY;

//    wsprintf(strTmp,"OpenTelnetConnection: con->hScreen= %lu \n\r",con->hScreen);
//    OutputDebugString(strTmp);

//    wsprintf(strTmp,"OpenTelnetConnection: NumConnections= %d \n\r",NumConnections);
//    OutputDebugString(strTmp);
    s=(SOCKET) socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    con->socket=(SOCKET)s;
    sockaddr.sin_family = AF_INET ;  
    sockaddr.sin_addr.s_addr = htonl( INADDR_ANY );
    sockaddr.sin_port = htons( 0 ) ;
//    MessageBox(NULL,"About to try bind...","wintel",MB_OK|MB_ICONEXCLAMATION);
    ret=bind(con->socket,(struct sockaddr far *)&sockaddr,(int)sizeof(struct sockaddr_in));
    if (ret==SOCKET_ERROR) {
        DeleteConnectionFromList(s);        
        wsprintf(buf,"1) Socket error = %d\r\n", WSAGetLastError());
        OutputDebugString(buf);
        wsprintf(buf,"Socket error on bind!");
        MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr==NULL) OutputDebugString("fpScr is NULL! \r\n");
        DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ConnectionList);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    }
    remote_host = gethostbyname(tmpConfig->title);
    if (!remote_host) {
        DeleteConnectionFromList(s);        
        wsprintf(buf,"Unable to lookup %s!",tmpConfig->title);
        OutputDebugString(buf);
        MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr==NULL) OutputDebugString("fpScr is NULL! \r\n");
        DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        DeleteConnectionFromList(s);
        GlobalUnlock(ConnectionList);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    }        
    remote_addr.sin_family= AF_INET;    
    remote_addr.sin_addr.S_un.S_un_b.s_b1= remote_host->h_addr[0];
    remote_addr.sin_addr.S_un.S_un_b.s_b2= remote_host->h_addr[1];
    remote_addr.sin_addr.S_un.S_un_b.s_b3= remote_host->h_addr[2];
    remote_addr.sin_addr.S_un.S_un_b.s_b4= remote_host->h_addr[3];
    remote_addr.sin_port= htons(23);
    if(connect(con->socket,(struct sockaddr far*) &remote_addr, sizeof(struct sockaddr))){
        DeleteConnectionFromList(s);    
        OutputDebugString("Connect error for s \r\n");
        MessageBox(NULL,"Connect error!",NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr==NULL) OutputDebugString("fpScr is NULL! \r\n");
        DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ConnectionList);
        GlobalUnlock(hGlobalMem);        
        GlobalFree(hGlobalMem);
        return(-1);
    }
    start_negotiation(con->socket);
//    con->telstate=OPENED;
    con->telstate=0;
    NumConnections++;
    LocalUnlock(hTitleMem);
    GlobalUnlock(hGlobalMem);
    GlobalUnlock(ConnectionList);
    return(1);
}

void DeleteConnectionFromList(SOCKET searchID) {
    HGLOBAL ghCon,ghTmp;
    CONNECTION far *fpCon,far *fpTmp;

    ghCon=ConnectionList;
    if (ghCon==NULL) return;
    
    while(ghCon) {
        fpCon = (CONNECTION far *)GlobalLock(ghCon);
        if (fpCon==NULL) return;
        if (fpCon->socket == searchID) { /* Found it! */
            if (ghCon==ConnectionList) { /* top element */
//                OutputDebugString("Delete top element \r\n");
                ConnectionList=fpCon->next;
            }                
            if (fpCon->next) {     
//                OutputDebugString("next != NULL \r\n");
                
                fpTmp=(CONNECTION far *)GlobalLock((HGLOBAL)fpCon->next);
                fpTmp->prev=fpCon->prev;
                GlobalUnlock(fpCon->next);
            }            
            if (fpCon->prev) {
//                OutputDebugString("prev != NULL \r\n");
                fpTmp=(CONNECTION far *)GlobalLock(fpCon->prev);
                fpTmp->next=fpCon->next;
                GlobalUnlock(fpCon->prev);
            }
            GlobalUnlock(ghCon);
            return;
        } else {  /* check next item */
            ghTmp=fpCon->next;
            GlobalUnlock(ghCon);
            ghCon=ghTmp;
        }    
    }        
}

int GetNewConnection(void) {
    HGLOBAL ghCon;
    CONNECTION far *fpCon,far *fpTmp;

    ghCon = GlobalAlloc(GHND,sizeof(CONNECTION));
    if (ghCon==NULL) return -1;
    fpCon = (CONNECTION far *)GlobalLock(ghCon);
    if (fpCon==NULL) return -1;
    fpCon->backspace=TRUE;
    if (ConnectionList==NULL) {
//        OutputDebugString("Connection List is Null! \r\n");
        fpCon->next=NULL;
        fpCon->prev=NULL;
        GlobalUnlock(ghCon);
        ConnectionList=ghCon;
        return(OPEN_STANDBY);
    } else {
//        OutputDebugString("Adding Element to front of list! \r\n");
        fpCon->next=ConnectionList;
        fpCon->prev=NULL;
        fpTmp = (CONNECTION far *)GlobalLock(ConnectionList);
        fpTmp->prev=ghCon;
        ConnectionList=ghCon;
        GlobalUnlock(fpCon->next);
    }  
    GlobalUnlock(ghCon);
    ConnectionList=ghCon;
    return(OPEN_STANDBY);
}
   
/****************************************************************************/
int NEAR DoDialog(char *szDialog, FARPROC lpfnDlgProc) {
    int nReturn;
    
    lpfnDlgProc= MakeProcInstance(lpfnDlgProc,hInst);
    if (lpfnDlgProc==NULL)
        MessageBox(hWnd,"Couldn't make procedure instance",NULL,MB_OK);    
    
    wsprintf(strTmp,"szDialog =%s hWnd=%lu hInst=%lu ptr= %lu \r\n",szDialog,hWnd,hInst,lpfnDlgProc);
//    OutputDebugString(strTmp);
    nReturn= DialogBox(hInst,szDialog,hWnd,lpfnDlgProc);
    if (nReturn==-1)
//        OutputDebugString("nReturn=-1 \r\n");
    FreeProcInstance(lpfnDlgProc);
    return (nReturn);
}    
/****************************************************************************

    FUNCTION: About(HWND, unsigned, WORD, LONG)

    PURPOSE:  Processes messages for "About" dialog box

    MESSAGES:

    WM_INITDIALOG - initialize dialog box
    WM_COMMAND    - Input received

    COMMENTS:

    No initialization is needed for this particular dialog box, but TRUE
    must be returned to Windows.

    Wait for user to click on "Ok" button, then close the dialog box.

****************************************************************************/

BOOL FAR PASCAL About(hDlg, message, wParam, lParam)
HWND hDlg;                                /* window handle of the dialog box */
WORD message;                         /* type of message                 */
WORD wParam;                              /* message-specific information    */
LONG lParam;
{
    switch (message) {
    case WM_INITDIALOG:        /* message: initialize dialog box */
        return (TRUE);

    case WM_COMMAND:              /* message: received a command */
        if (wParam == IDOK                /* "OK" box selected?      */
                || wParam == IDCANCEL) {      /* System menu close command? */
        EndDialog(hDlg, TRUE);        /* Exits the dialog box        */
        return (TRUE);
        }
        break;
    }
    return (FALSE);               /* Didn't process a message    */
}
/****************************************************************************

    FUNCTION: OpenTelnetDlg(HWND, unsigned, WORD, LONG)

    PURPOSE:  Processes messages for "Open New Telnet Connection" dialog box

    MESSAGES:

    WM_INITDIALOG - initialize dialog box
    WM_COMMAND    - Input received

****************************************************************************/

BOOL FAR PASCAL OpenTelnetDlg(HWND hDlg, WORD message, WORD wParam, LONG lParam) {
    char szConnectName[256];
    HDC  hDC;
    int  xExt,yExt;
    DWORD Ext;
    HWND hEdit;
    int  nLen;
    int iHostNum=1;
    char tmpName[128],tmpBuf[80];
    
    switch (message) {
        case WM_INITDIALOG:        /* message: initialize dialog box */
            hDC=GetDC(hDlg);
            Ext=GetDialogBaseUnits();
            xExt=(190 *LOWORD(Ext)) /4 ;
            yExt=(72 * HIWORD(Ext)) /8 ;
//            CheckRadioButton(hDlg,TEL_USEDEFAULTS,TEL_MANUALCONFIGURE,TEL_USEDEFAULTS);
//            SetDlgItemText(hDlg,TEL_CONNECT_NAME,"void");
            GetPrivateProfileString("hosts","host0","",tmpName,128,"ncsa_net.ini");
            if (tmpName[0]) {
            	*(_fstrchr(tmpName,','))='\0';  
            	SetDlgItemText(hDlg,TEL_CONNECT_NAME,tmpName);
            }	
            hEdit=GetWindow(GetDlgItem(hDlg,TEL_CONNECT_NAME),GW_CHILD);
            for(;;) {
                char far *tmpCommaLoc;
                wsprintf(tmpBuf,"host%d",iHostNum++);
                GetPrivateProfileString("hosts",tmpBuf,"",tmpName,128,"ncsa_net.ini");
                tmpCommaLoc=_fstrchr(tmpName,',');  
                if (tmpCommaLoc) *tmpCommaLoc='\0';
                if (tmpName[0])
                    SendDlgItemMessage(hDlg,TEL_CONNECT_NAME,CB_ADDSTRING,0,(LPARAM)((LPSTR)tmpName));
                else break; 
            }
//            SendDlgItemMessage(hDlg,TEL_CONNECT_NAME,CB_ADDSTRING,0,(LPARAM)((LPSTR)"Void"));
//            SendDlgItemMessage(hDlg,TEL_CONNECT_NAME,CB_ADDSTRING,0,(LPARAM)((LPSTR)"x"));
//            SendDlgItemMessage(hDlg,TEL_CONNECT_NAME,CB_ADDSTRING,0,(LPARAM)((LPSTR)"Uxa"));
            SetWindowPos(hDlg,NULL,(GetSystemMetrics(SM_CXSCREEN)/2)-(xExt/2),
                (GetSystemMetrics(SM_CYSCREEN)/2)-(yExt/2),0,0,
                SWP_NOSIZE|SWP_NOZORDER|SWP_SHOWWINDOW);
            ReleaseDC(hDlg,hDC);      
            SendMessage(hEdit,WM_USER+1,NULL,NULL);
            SendMessage(hDlg,WM_SETFOCUS,NULL,NULL);
            return (TRUE);

        case WM_COMMAND:              /* message: received a command */
            switch (wParam) {
                case TEL_CANCEL:
                     EndDialog(hDlg, FALSE);        /* Exits the dialog box        */
                     break;
                 
                case TEL_OK:
                     GetDlgItemText(hDlg,TEL_CONNECT_NAME,szConnectName,256);
                     nLen= lstrlen(szConnectName);
                     if (nLen==0) { 
                        MessageBox(hDlg,"You must enter a session name!",NULL,MB_OK);
                        break;
                        }
                     hTitleMem= GlobalAlloc(GPTR,nLen);
                     tmpConfig->title= (char far *) GlobalLock(hTitleMem);
                     lstrcpy(tmpConfig->title,szConnectName);
                     tmpConfig->width=DEF_WIDTH;
                     tmpConfig->height=DEF_HEIGHT;                            
                     EndDialog(hDlg,TRUE);
                     break;
            }                
            return (FALSE);                    
    }
    return (FALSE);               /* Didn't process a message    */
}

BOOL FAR PASCAL ConfigSessionDlg(hDlg, message, wParam, lParam)
HWND hDlg;                                /* window handle of the dialog box */
WORD message;                         /* type of message                 */
WORD wParam;                              /* message-specific information    */
LONG lParam;
{
    char szWindowTitle[256];
    HDC  hDC;
    int  xExt,yExt;
    DWORD Ext;
    int  nLen;
    
    switch (message) {
        case WM_INITDIALOG:        /* message: initialize dialog box */
//            OutputDebugString("Begining! \r\n");
            
            hDC=GetDC(hDlg);
            Ext=GetDialogBaseUnits();
            xExt=(160 *LOWORD(Ext)) /4 ;
            yExt=(130 * HIWORD(Ext)) /8 ;
            wsprintf(strTmp,"xExt= %d yExt=%d \r\n",xExt,yExt);
//            OutputDebugString(strTmp);
            CheckRadioButton(hDlg,CON_COLUMNS132,CON_COLUMNS80,CON_COLUMNS80);
            CheckRadioButton(hDlg,CON_BACKSPACE,CON_DELETE,CON_BACKSPACE);
            CheckRadioButton(hDlg,CON_CRLF,CON_CRNUL,CON_CRLF);
            CheckRadioButton(hDlg,CON_BUFFERS,CON_SENDS,CON_BUFFERS);
#ifdef NOT
            CheckDlgButton(hDlg,CON_SCRLBCK,TRUE);
            SetDlgItemText(hDlg,CON_NUMLINES,"200");
#endif
            SetDlgItemText(hDlg,CON_SESSIONNAME,tmpConfig->title);
            SetDlgItemText(hDlg,CON_WINDOWTITLE,tmpConfig->title);
            SetWindowPos(hDlg,NULL,(GetSystemMetrics(SM_CXSCREEN)/2)-(xExt/2),
                (GetSystemMetrics(SM_CYSCREEN)/2)-(yExt/2),0,0,
                SWP_NOSIZE|SWP_NOZORDER|SWP_SHOWWINDOW);
            ReleaseDC(hDlg,hDC);     
            return (TRUE);

        case WM_COMMAND:              /* message: received a command */
            switch (wParam) {
                case CON_USEDEFAULTS:
                     EndDialog(hDlg, FALSE);        /* Exits the dialog box        */
                     break;
                 
                case CON_OK:
                     GetDlgItemText(hDlg,CON_WINDOWTITLE,szWindowTitle,256);
                     nLen= lstrlen(szWindowTitle);
                     if (nLen) {
                        hTitleMem= GlobalAlloc(GPTR,nLen);
                        tmpConfig->title= (char far *) GlobalLock(hTitleMem);
                        lstrcpy(tmpConfig->title,szWindowTitle);
                        }
                     if (IsDlgButtonChecked(hDlg,CON_COLUMNS132)) tmpConfig->width=132;
                     else tmpConfig->width=DEF_WIDTH;
                     tmpConfig->height=DEF_HEIGHT;                            
                         
                     EndDialog(hDlg,TRUE);
                     break;
            }                
            return (FALSE);                    
    }
    return (FALSE);               /* Didn't process a message    */
}

